/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * This is GNU GO, a Go program. Contact gnugo@gnu.org, or see   *
 * http://www.gnu.org/software/gnugo/ for more information.      *
 *                                                               *
 * Copyright 1999 by the Free Software Foundation.               *
 *                                                               *
 * This program is free software; you can redistribute it and/or *
 * modify it under the terms of the GNU General Public License   *
 * as published by the Free Software Foundation - version 2.     *
 *                                                               *
 * This program is distributed in the hope that it will be       *
 * useful, but WITHOUT ANY WARRANTY; without even the implied    *
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR       *
 * PURPOSE.  See the GNU General Public License in file COPYING  *
 * for more details.                                             *
 *                                                               *
 * You should have received a copy of the GNU General Public     *
 * License along with this program; if not, write to the Free    *
 * Software Foundation, Inc., 59 Temple Place - Suite 330,       *
 * Boston, MA 02111, USA                                         *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */





#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <assert.h>
#include "liberty.h"



/* ================================================================ */
/*                 Pushing and popping of boards                    */
/* ================================================================ */


/* Stack of boards for storing positions while reading. */
static board_t  stack[MAXSTACK][MAX_BOARD][MAX_BOARD]; 

/* Stack of trial moves to get to current position 
 * position and which color made them. Perhaps 
 * this should be one array of a structure 
 */
static int      stacki[MAXSTACK];
static int      stackj[MAXSTACK];      
static int      move_color[MAXSTACK];  
static int      ko_stacki[MAXSTACK];
static int      ko_stackj[MAXSTACK];      

static Hash_data  hashdata_stack[MAXSTACK];

/* Stack of black and white captured stones */
static int      stackbc[MAXSTACK];     
static int      stackwc[MAXSTACK];


/*
 * pushgo pushes the position onto the stack.  Returns the stack
 * pointer after the push.
 */

int 
pushgo()
{
  if (showstack)
    gprintf("        *** STACK before push: %d\n", stackp); 
  memcpy(stack[stackp], p, sizeof(p));
  ko_stacki[stackp] = ko_i;
  ko_stackj[stackp] = ko_j;
  memcpy(&hashdata_stack[stackp], &hashdata, sizeof(hashdata));

  stackbc[stackp]=black_captured;
  stackwc[stackp]=white_captured;

  return ++stackp;
}


/*
 * popgo pops the position from the stack.  Returns the stack
 * pointer after the pop.
 */

int 
popgo()
{
  stackp--;
  if (showstack)
    gprintf("<=    *** STACK  after pop: %d\n", stackp);
   
  memcpy(p, stack[stackp], sizeof(p) );
  ko_i = ko_stacki[stackp];
  ko_j = ko_stackj[stackp];

  memcpy(&hashdata, &(hashdata_stack[stackp]), sizeof(hashdata) );

  black_captured=stackbc[stackp];
  white_captured=stackwc[stackp];
  if (count_variations) 
    sgf_write_line(")\n");

  return stackp;
}


/*
 * trymove pushes the position onto the stack, and makes a move
 * at (i, j) of color. Returns one if the move is legal. The
 * stack pointer is only incremented if the move is legal.
 *
 * So the way to use this is:
 *
 *   if (trymove(i, j, color, [message], k, l)) {
 *      ...
 *      popgo();
 *   }   
 *
 * The message can be written as a comment to an sgf file using 
 * sgfdump().  (k, l) can be -1 if they are not needed but if they are 
 * not the location of (k, l) is included in the comment.
 */

int 
trymove(int i, int j, int color, const char *message, int k, int l)
{
  /* We could call legal() here, but that has to do
   * and then un-do all the stuff we want to re-do here.
   * (pushgo, updateboard, ...)  So we replicate its functionality
   * inline.
   */

  /* 1. The move must be inside the board or i == -1 (pass) */
  assert (i>=0 && i<board_size && j>=0 && j<board_size);

  /* 2. The location must be empty. */
  if (p[i][j]!=EMPTY)
    return 0;

  /* 3. The location must not be the ko point. */
  if (i == ko_i && j == ko_j)
    if (((i>0) && (p[i-1][j] != color))
	|| ((i==0) && (p[i+1][j] != color)))
    {
      RTRACE("%m would violate the ko rule\n", i, j);
      return 0;
    }

  /* So far, so good.  Now push the move on the move stack. */
  stacki[stackp]=i;
  stackj[stackp]=j;
  move_color[stackp]=color;

  /* Check for stack overflow. */
  if (stackp >= MAXSTACK-2) {
    fprintf(stderr, 
	    "gnugo: Truncating search. This is beyond my reading ability!\n");
    return 0;
  }

  /* Ok, the trivial tests went well. Push the position on the stack, 
     and try to actually make the move. */
  pushgo();

  if (verbose==4)
    dump_stack();

  if (updateboard(i,j,color) == 0) {

    /* no trivial liberties : we need to run approxlib to
     * make sure we do actually have some liberties !
     */

    if (approxlib(i,j,color,1) == 0) {
      RTRACE("%m would be suicide\n", i, j);

      /* Almost do the work of popgo().  The only difference is that
	 we don't close the variations in the sgf file. */
      stackp--;
      if (showstack)
	gprintf("<=    *** STACK  after pop: %d\n", stackp);
   
      memcpy(p, stack[stackp], sizeof(p) );
      ko_i = ko_stacki[stackp];
      ko_j = ko_stackj[stackp];
      memcpy(&hashdata, &(hashdata_stack[stackp]), sizeof(hashdata) );
      black_captured = stackbc[stackp];
      white_captured = stackwc[stackp];

      return 0;
    }
  }
  /*  hashdata_set_tomove(&hashdata, OTHER_COLOR(color));*/

  /* Store the move in the sgf file if there is a dump file. */
  if (sgf_dump) {
    if (k == -1) {
      if (color==BLACK)
	sgf_write_line("\n(;B[%c%c]C[%s (%d)\n]", 
		       'a'+j, 'a'+i, message, count_variations);
      if (color==WHITE)
	sgf_write_line("\n(;W[%c%c]C[%s (%d)\n]",
		       'a'+j, 'a'+i, message, count_variations);
    } else {
      char movename[4];
	
      if (l < 8)
	*movename=l+65;
      else
	*movename=l+66;
      sprintf(movename+1, "%d", board_size-k);
	
      if (color==BLACK)
	sgf_write_line("\n(;B[%c%c]C[%s at %s (%d)\n]",
		       'a'+j, 'a'+i, message, movename, count_variations);
      if (color==WHITE)
	sgf_write_line("\n(;W[%c%c]C[%s at %s (%d)\n]",
		       'a'+j, 'a'+i, message, movename, count_variations);
    }
  }

  if (count_variations)
    count_variations++;
  stats.nodes++;

  return 1;
}


/*
 * tryko pushes the position onto the stack, and makes a move
 * at (i, j) of color. The move is allowed even if it is an
 * illegal ko capture. It is to be imagined that (color) has
 * made an intervening ko threat which was answered and now
 * the continuation is to be explored.
 *
 * Return 1 if the move is legal with the above caveat.
 */

int 
tryko(int i, int j, int color, const char *message)
{
  assert (i>=0 && i<board_size && j>=0 && j<board_size);

  if (p[i][j]!=EMPTY)
    return 0;

  stacki[stackp]=i;
  stackj[stackp]=j;
  move_color[stackp]=color;

  if (stackp >= MAXSTACK-2) {
    fprintf(stderr, 
	    "gnugo: Truncating search. This is beyond my reading ability!\n");
    return 0;
  }

  pushgo();

  if (verbose==4)
    dump_stack();

  if (updateboard(i,j,color) == 0) {
    fprintf(stderr, "tryko: illegal move\n");
    verbose=4;
    dump_stack();
    abort;
  }

  if (sgf_dump) {
    if (color==BLACK)
      sgf_write_line("\n(;B[tt]C[tenuki (ko threat)]\n\
;W[tt]C[tenuki (answers ko threat)]\n;B[%c%c]C[%s (%d)\n]", 
		     'a'+j, 'a'+i, message, count_variations);
    else if (color==WHITE)
      sgf_write_line("\n(;W[tt]C[tenuki (ko threat)]\n\
;B[tt]C[tenuki (answers ko threat)]\n;B[%c%c]C[%s (%d)\n]",
		       'a'+j, 'a'+i, message, count_variations);
  }
  if (count_variations)
    count_variations++;
  stats.nodes++;

  return 1;
}


/*
 * dump_stack() for use under gdb prints the move stack. 
 */

void
dump_stack(void)
{
  int n;

  for (n=0; n<stackp; n++) {
    gprintf("%o%s:%m ", move_color[n] == BLACK ? "B" : "W",
	    stacki[n], stackj[n]);
  }

#if !TRACE_READ_RESULTS
  if (count_variations)
    gprintf("%o (variation %d)", count_variations-1);
#else
  gprintf("%o (%d)", hashdata.hashval);
#endif

  gprintf("%o\n");
}


/* ================================================================ */
/*                         Board manipulation                       */
/* ================================================================ */


/*
 * do_add_stone and do_remove_stone are defined as macros for 
 * efficiency reasons.
 */

#define do_add_stone(i, j, color) \
  do { \
    p[i][j]=color; \
    hashdata_invert_stone(&hashdata, i, j, color); \
  } while (0)

#define do_remove_stone(i, j) \
  do { \
    hashdata_invert_stone(&hashdata, i, j, p[i][j]); \
    p[i][j] = EMPTY; \
  } while (0)


/*
 * And now the functions, accessible from outside this file.
 */

void
add_stone(int i, int j, int color)
{
  assert(p[i][j] == EMPTY);
  do_add_stone(i, j, color);
}


void
remove_stone(int i, int j)
{
  assert(p[i][j] != EMPTY);
  do_remove_stone(i, j);
}


/*
 * Remove the string at (i,j), treating it as captured.
 * FIXME: Move to utils.c since this isn't a basic function?
 */

void
remove_string(int i, int j)
{
  int color=p[i][j];

  assert(color != EMPTY);

  do_remove_stone(i, j);
  if (color==BLACK)
    black_captured++;
  else
    white_captured++;
  if ((i>0) && (p[i-1][j]==color))
    remove_string(i-1, j);
  if ((i<board_size-1) && (p[i+1][j]==color))
    remove_string(i+1, j);
  if ((j>0) && (p[i][j-1]==color))
    remove_string(i, j-1);
  if ((j<board_size-1) && (p[i][j+1]==color))
    remove_string(i, j+1);
}


/* ---------------------------------------------------------------- */


static int check_for_capture(int m, int n, int color);

/*
 * Place a "color" on the board at i,j, and remove
 * any captured stones.  If (i, j) == (boardsize, boardsize),
 * the move is a pass, but that is OK too.
 *
 * Assumes that the only stones which can be captured are
 * those immediately adjacent to the piece just played.
 *
 * Returns non-zero if there are any "trivial" liberties, ie if
 * we capture any stones or there are any adjacent liberties.
 * A return value of 0 does not mean that the placed stone has no
 * liberties, though a return of >0 means it definately does.
 * (ie return of >0 is sufficient, though not necessary, test
 * for move not being a suicide.)
 *
 *  Globals affected : uses count(), so anything it affects
 */

int
updateboard(int i, int j, int color)
{
  int other = OTHER_COLOR(color);
  int obvious_liberties=0;
  int captures = 0;
  int tot_captures = 0;

#if CHECK_HASHING
  {
    Hashposition pos;
    unsigned long hash;

    /* Check the hash table to see if we have had this position before. */
    board_to_position(p, ko_i, ko_j, /*color,*/ &pos);
    hash = board_hash(p, ko_i, ko_j/*, color*/);
    assert(hash == hashdata.hashval);
    assert(hashposition_compare(&pos, &(hashdata.hashpos)) == 0);
  }
#endif

  last_move_i=i;
  last_move_j=j;

  ko_i = -1;
  ko_j = -1;
  hashdata_remove_ko(&hashdata);

  /* If the move is a pass, we are done here. */
  if (i >= board_size && j >= board_size)
    return -1;

  assert(i>=0 && i<board_size && j>=0 && j<board_size);
  ASSERT( p[i][j] == EMPTY, i, j);

  if (stackp==0) 
    color_has_played|=color;

  do_add_stone(i, j, color);
  DEBUG(DEBUG_BOARD, "Update board: %m = %d\n", i,j, color);

  /* Check to the top. */
  if (i>0) {
    if (p[i-1][j] == other) {
      tot_captures = check_for_capture(i-1, j, other);
      obvious_liberties = tot_captures;
    } else if (p[i-1][j] == EMPTY)
      ++obvious_liberties;
  }

  /* Check to the bottom. */
  if (i<board_size-1) {
    if (p[i+1][j]==other) {
      captures = check_for_capture(i+1, j, other);
      tot_captures += captures;
      obvious_liberties += captures;
    } else if (p[i+1][j] == EMPTY)
      ++obvious_liberties;
  }

  /* Check to the left. */
  if (j>0) {
    if (p[i][j-1]==other) {
      captures = check_for_capture(i, j-1, other);
      obvious_liberties += captures;
      tot_captures += captures;
    } else if (p[i][j-1]==EMPTY)
      ++obvious_liberties;
  }

  /* Check to the right. */
  if (j<board_size-1) {
    if (p[i][j+1]==other) {
      captures = check_for_capture(i, j+1, other);
      obvious_liberties += captures;
      tot_captures += captures;
    } else if (p[i][j+1]==EMPTY)
      ++obvious_liberties;
  }

  /* Check for possible ko and store it into the ko variables. */
  if (tot_captures == 1) {
    int  friends = 0;
    int  libs = 0;
    int  libertyi = -1;
    int  libertyj = -1;

    if (i > 0) {
      if (p[i-1][j] == color) 
	friends++;
      if (p[i-1][j] == EMPTY) {
	libs++;
	libertyi = i - 1;
	libertyj = j;
      }
    }
    if (i < board_size - 1) {
      if (p[i+1][j] == color) 
	friends++;
      if (p[i+1][j] == EMPTY) {
	libs++;
	libertyi = i + 1;
	libertyj = j;
      }
    }
    if (j > 0) {
      if (p[i][j-1] == color) 
	friends++;
      if (p[i][j-1] == EMPTY) {
	libs++;
	libertyi = i;
	libertyj = j - 1;
      }
    }
    if (j < board_size - 1) {
      if (p[i][j+1] == color) 
	friends++;
      if (p[i][j+1] == EMPTY) {
	libs++;
	libertyi = i;
	libertyj = j + 1;

      }
    }

    /* This test guarantees that the newly put stone has exactly one
     * liberty (the newly caught stone) and that it doesn't connect
     * to any friendly stones.  In the latter case it is called 
     * a snapback, and not a ko, which is ok.  
     */
    if (libs == 1 && friends == 0) {
      ko_i = libertyi;
      ko_j = libertyj;
      hashdata_set_ko(&hashdata, libertyi, libertyj);
    }
  }

#if CHECK_HASHING
  {
    Hashposition pos;
    unsigned long hash;

    /* Check the hash table to see if we have had this position before. */
    board_to_position(p, ko_i, ko_j, /*color,*/ &pos);
    hash = board_hash(p, ko_i, ko_j/*, color*/);
    assert(hashposition_compare(&pos, &(hashdata.hashpos)) == 0);
    assert(hash == hashdata.hashval);
  }
#endif

  return obvious_liberties;
}


/*
 * legal(i, j, color) determines whether the move (color) at
 * (i, j) is legal.
 */

int 
legal(int i, int j, int color)
{
  int legit=0;

  /* 0. A pass move is always legal. */
  if (i == board_size && j == board_size)
    return 1;

  /* 1. The move must be inside the board. */
  assert(i>=0 && i<board_size && j>=0 && j<board_size);

  /* 2. The location must be empty. */
  if (p[i][j]!=EMPTY) 
    return 0;

  /* 3. The location must not be the ko point. */
  if (i == ko_i && j == ko_j)
    if (((i>0) && (p[i-1][j] != color))
	|| ((i==0) && (p[i+1][j] != color)))
    {
      RTRACE("%m would violate the ko rule\n", i, j);
      return 0;
    }

  /* Check for stack overflow. */
  if (stackp >= MAXSTACK-2) {
    fprintf(stderr, 
	    "gnugo: Truncating search. This is beyond my reading ability!\n");
    return 0;
  }

  /* Ok, the trivial tests went well. Push the position on the stack, 
     and try to actually make the move. */
  pushgo();
  updateboard(i, j, color);

  /* Only care about 0 or >= 1. */
  if (approxlib(i, j, color, 1)==0) {
    RTRACE("%m would be suicide\n", i, j);
    legit = 0;
  } else
    legit = 1;

  stackp--;
  if (showstack)
    gprintf("<=    *** STACK  after pop: %d\n", stackp);
   
  memcpy(p, stack[stackp], sizeof(p) );
  ko_i = ko_stacki[stackp];
  ko_j = ko_stackj[stackp];
  memcpy(&hashdata, &(hashdata_stack[stackp]), sizeof(hashdata) );
  black_captured=stackbc[stackp];
  white_captured=stackwc[stackp];

  return legit;
}


/* ---------------------------------------------------------------- */


/* save_state() and restore_state() backup and restore the board
 * position without using the stack. They are used in counting
 * under the ascii interface by endgame().
 */

static board_t p_backup[MAX_BOARD][MAX_BOARD];
static int black_captured_backup;
static int white_captured_backup;


void
save_state()
{
  black_captured_backup=black_captured;
  white_captured_backup=white_captured;
  memcpy(p_backup, p, sizeof(p));
}


void
restore_state()
{
  black_captured=black_captured_backup;
  white_captured=white_captured_backup;
  memcpy(p, p_backup, sizeof(p));
}


/* ---------------------------------------------------------------- */


/*
 * If the chain at (m,n) has no liberties, remove it from
 * the board. Return the number of stones captured.
 */

static char check_mx[MAX_BOARD][MAX_BOARD];
static char check_mark = -1;

static void
init_check_for_capture(void)
{
  /* set all pieces as unmarked */
  memset(check_mx, 0, sizeof(check_mx));
  check_mark = 1;
}


static int 
check_for_capture(int m, int n, int color)
{
  int i,j;
  int captured=0; /* number of captured */

  ASSERT(p[m][n] == color, m,n);

  DEBUG(DEBUG_COUNT,"Checking %m for capture\n", m,n);

  check_mark++;
  if (check_mark == 0) {
    /* We have wrapped around, reinitialize. */
    init_check_for_capture();
  }

  if (count(m, n, color, check_mx, 1, check_mark) > 0)
    return 0;

  /* Remove all the stones with mx[i][j] == 0,
   *  (since these are the ones count() touched)
   * - count also marks EMPTY stones, but there's
   * no harm in resetting them to EMPTY
   */

  for (i=0; i<board_size; ++i)
    for (j=0; j<board_size; ++j)
      if (check_mx[i][j] == check_mark) {
	/* FIXME: We should really not remove the already removed stones.
	          This seriously upsets the hash code. /iw */
	do_remove_stone(i, j);
	++captured;
      }
  
  if (color == WHITE)
    white_captured += captured;
  else
    black_captured += captured;

  return captured;
}


/*
 * Initialize board.
 */

void
init_board()
{
  memset(p, EMPTY, sizeof(p));
  ko_i = -1;
  ko_j = -1;
  hashdata_init(&hashdata);
  last_move_i=last_move_j=-1;
}


/*
 * Initialize the gnugo engine.
 *
 * FIXME: This function doesn't belong here. Move it?
 */

void
init_gnugo(int memory)
{
  /* FIXME: The following lines don't actually initialize the board,
     but the engine. */
  depth=DEPTH;
  backfill_depth=BACKFILL_DEPTH;
  fourlib_depth=FOURLIB_DEPTH;
  ko_depth=KO_DEPTH;
  
#if HASHING
  /* Initialize hash table.
   *
   * The number 1.4 below is the quotient between the number of nodes
   * and the number of read results.  It was found in a test that this 
   * number varies between 1.15 and 1.4.  Thus we use 1.4.
   */
  {
    float nodes;
    
    nodes = ((float) (memory * 1024 * 1024)
	     / (sizeof(Hashnode) + sizeof(Read_result) * 1.4));
    /*                       table size,  nodes, read results */
    movehash = hashtable_new(1.5 * nodes, nodes, nodes * 1.4);
  } 
#endif
  depth=DEPTH;
  backfill_depth=BACKFILL_DEPTH;
  fourlib_depth=FOURLIB_DEPTH;
  ko_depth=KO_DEPTH;
  
#if HASHING
  /* Initialize hash table.
   *
   * The number 1.4 below is the quotient between the number of nodes
   * and the number of read results.  It was found in a test that this 
   * number varies between 1.15 and 1.4.  Thus we use 1.4.
   */
  {
    float nodes;
    
    nodes = ((float) (memory * 1024 * 1024)
	     / (sizeof(Hashnode) + sizeof(Read_result) * 1.4));
    /*                       table size,  nodes, read results */
    movehash = hashtable_new(1.5 * nodes, nodes, nodes * 1.4);
  } 
#endif

  init_board();
}


/*
 * Count (minimum) liberty of color piece at (m, n).
 * m,n may contain either that color, or be empty, in which
 * case we get what the liberties would be if we were to place
 * a piece there.
 * If the caller only cares about a minimum liberty count, it
 * should pass that minimum in, and then we *may* stop counting
 * if/when we reach that number. (We may still return a larger
 * number if we feel like it. (The current implementation doesn't.))
 *
 * Note that countlib(i,j,c) is defined as approxlib(i,j,c,99999)
 */

static char approxlib_ml[MAX_BOARD][MAX_BOARD];
static char approxlib_mark = -1;

static void
init_approxlib(void)
{
  /* Set all pieces as unmarked. */
  memset(approxlib_ml,0,sizeof(approxlib_ml));
  approxlib_mark = 1;
}
  

int
approxlib(int m,      /* row number 0 to board_size-1 */
	  int n,      /* column number 0 to board_size-1 */
	  int color,  /* BLACK or WHITE */
	  int maxlib) /* count might stop when it reaches this */
{
  int i;

  ASSERT(m>=0 && m < board_size && n >= 0 && n < board_size,m,n);
  ASSERT(color != EMPTY,m,n);
  
  ASSERT(p[m][n] != OTHER_COLOR(color), m,n);

  approxlib_mark++;
  if (approxlib_mark == 0) { 
    /* We have wrapped around, reinitialize. */
    init_approxlib();
  }

  /* Count liberty of current piece. */
  i = count(m, n, color, approxlib_ml, maxlib, approxlib_mark);

  /* trying to get rid of globals... */
  assert(i == lib);  /* return value should match global in this case */
  
  DEBUG(DEBUG_COUNT, "approxlib(%m,%d) returns %d\n", m,n,maxlib,i);
  
  return i;
}


/*
 * Count (minimum) liberty of color piece at location (i, j)
 * and return value in global variable lib. Return size of
 * connected component in size.
 *
 * No check is made on entry that that square is occupied
 * by that color; It is valid to use this to count the
 * liberties that there would be if color was put into
 * empty square at (i,j).
 *
 * If k<lib, then libi[k],libj[k] points to one of the
 * liberties of the string.
 *
 * ENTRY: mark is the current value of "mark"
 *
 *        mx[][] contains "mark" for stones and empty intersections that
 *	  may not be counted. Typically "mark" is a value that at calling
 *	  time is absent from mx[][].
 *
 *        maxlib : typically, it is not necessary to get the
 *        exact number of liberties : we can stop when it
 *        reaches some threshold (often 1 !).
 *
 * RETURNS : number of liberties found
 *
 * EXIT :  return num of liberties in global "lib"
 *
 *        return positions of liberties in globals libi[] and libj[]
 *
 *        return group size in global "size"
 *
 *        updates mx[][], putting "mark" at all stones and empty
 *	  intersections visited
 *
 *	  EITHER : return < maxlib, and mx[][] and l[][] fully updated
 *	  OR     : return == maxlib, and mx[][] and l[][] partially updated
 *
 * NOTE: intended as an internal helper for approxlib, but
 *       is exported as a general (though complex) function
 *       for efficiency.
 */

static int count_stacki[MAXCHAIN];
static int count_stackj[MAXCHAIN];

int
count(int i,     /* row number 0 to board_size-1 */
      int j,     /* column number 0 to board_size-1 */
      int color, /* BLACK or WHITE */
      char mx[MAX_BOARD][MAX_BOARD],  /* workspace: mx[][]=mark means stone already visited. */
      int maxlib,       /* Often, we only care if liberties > some minimum (usually 1) */
      char mark) /* Value to mark visited points with. */
{
  int count_stackp;
  
  mx[i][j] = mark;
  count_stacki[0] = i;
  count_stackj[0] = j;
  count_stackp = 1;
  size = 0;
  lib = 0;
  
  while (count_stackp>0) {
    count_stackp--;
    i = count_stacki[count_stackp];
    j = count_stackj[count_stackp];
    size++;

    /* check West neighbor */
    if (j != 0 && mx[i][j-1] != mark) {
      if (p[i][j-1] == EMPTY) {
	mx[i][j-1] = mark;
	libi[lib] = i;
	libj[lib] = j-1;
	++lib;
	if (lib == maxlib)
	  return lib;
      }
      else if (p[i][j-1] == color) {
	mx[i][j-1] = mark;
	count_stacki[count_stackp] = i;
	count_stackj[count_stackp] = j-1;
	count_stackp++;
      }
    }
    
    while(1) {
      /* check North neighbor */
      if (i != 0 && mx[i-1][j] != mark) {
	if (p[i-1][j] == EMPTY) {
	  mx[i-1][j] = mark;
	  libi[lib] = i-1;
	  libj[lib] = j;
	  ++lib;
	  if (lib == maxlib)
	    return lib;
	} 
	else if (p[i-1][j] == color) {
	  mx[i-1][j] = mark;
	  count_stacki[count_stackp] = i-1;
	  count_stackj[count_stackp] = j;
	  count_stackp++;
	}
      }

      /* check South neighbor */
      if (i != board_size-1 && mx[i+1][j] != mark) {
	if (p[i+1][j] == EMPTY) {
	  mx[i+1][j] = mark;
	  libi[lib] = i+1;
	  libj[lib] = j;
	  ++lib;
	  if (lib == maxlib)
	    return lib;
	}
	else if (p[i+1][j] == color) {
	  mx[i+1][j] = mark;
	  count_stacki[count_stackp] = i+1;
	  count_stackj[count_stackp] = j;
	  count_stackp++;
	}
      }

      /* check East neighbor */
      if (j == board_size-1 || mx[i][j+1] == mark)
	break;
      if (p[i][j+1] == EMPTY) {
	mx[i][j+1] = mark;
	libi[lib] = i;
	libj[lib] = j+1;
	++lib;
	if (lib == maxlib)
	  return lib;
	break;
      }
      else if (p[i][j+1] == color) {
	mx[i][j+1] = mark;
	size++;
	j++;
      }
      else
	break;
    }
  }

  return lib;
}


/*
 * Return true if the move (i,j) by (color) is a ko capture
 * (whether capture is legal on this move or not).
 */

int
is_ko(int i, int j, int color)
{
  int other=OTHER_COLOR(color);
  int captures=0;

  if (i>0) {
    if (p[i-1][j] != other)
      return 0;
    if (approxlib(i-1, j, other, 2)==1) {
      if (!singleton(i-1, j))
	return 0;
      captures++;
    }
  }

  if (i<board_size-1) {
    if (p[i+1][j] != other)
      return 0;
    if (approxlib(i+1, j, other, 2)==1) {
      if (!singleton(i+1, j))
	return 0;
      captures++;
    }
  }
  if (j>0) {
    if (p[i][j-1] != other) 
      return 0;
    if (approxlib(i, j-1, other, 2)==1) {
      if (!singleton(i, j-1))
	return 0;
      captures++;
    }
  }
  if (j<board_size-1) {
    if (p[i][j+1] != other)
      return 0;
    if (approxlib(i, j+1,other, 2)==1) {
      if (!singleton(i, j+1))
	return 0;
      captures++;
    }
  }

  return (captures==1);
}



/*
 * Local Variables:
 * tab-width: 8
 * c-basic-offset: 2
 * End:
 */
